#!/usr/bin/env python3
"""
This is python parser for the SWF format based on pandas' dataframe

Standard Workflow Format specification can be found at https://www.cs.huji.ac.il/labs/parallel/workload/swf.html.

[1] Chapin, S.J. et al. (1999). Benchmarks and Standards for the Evaluation of Parallel Job Schedulers.
In: Feitelson, D.G., Rudolph, L. (eds) Job Scheduling Strategies for Parallel Processing. JSSPP 1999.
Lecture Notes in Computer Science, vol 1659. Springer, Berlin, Heidelberg.

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2023
"""

import pandas as pd
import datetime

def parse_time(value) -> datetime.datetime:
    """
    This function parses a date as given by the output of "date" program in Linux.
    :param value: string with the date
    :return: datetime object with the parsed date
    """
    try:
        return datetime.datetime.strptime(value, "%a %b %d %H:%M:%S %Z %Y")
    except ValueError:
        print(f"[WARNING] {value} failed to create valid datetime object for start time. Ignoring timezone name.")
        split_value = value.split()
        if len(split_value) == 6:
            value = " ".join(split_value[:4] + [split_value[5]])  # remove timezone name
            try:
                return datetime.datetime.strptime(value, "%a %b %d %H:%M:%S %Y")
            except ValueError:
                print(f"[WARNING] {value} is not in the expected date format without timezone (e.g. Mon Nov 17 14:01:47 2008). Ignoring it.")
        else:
            print(f"[WARNING] {value} is not in the expected date format for start time (e.g. Mon Nov 17 14:01:47 CET 2008). Ignoring it.")


class SWFtrace:
    def __init__(self, swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                 unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime, max_memory,
                 allow_overuse, max_queues, queues, queue, max_partitions, partitions, partition, note, job_list,
                 file_location):
        # HEADER PARSER
        self._swf_version = swf_version
        self._computer = computer_name
        self._location = location
        self._acknowledge = acknowledge
        self._conversion = conversion
        self._max_jobs = max_jobs
        self._max_records = max_records
        self._preemption = preemption
        self._unix_start_time = unix_start_time
        self._time_zone_string = time_zone_string
        self._start_time = start_time
        self._end_time = end_time
        self._max_nodes = max_nodes
        self._max_procs = max_procs
        self._max_runtime = max_runtime
        self._max_memory = max_memory
        self._allow_overuse = allow_overuse
        self._max_queues = max_queues
        self._queues = queues
        self._queue = queue
        self._max_partitions = max_partitions
        self._partitions = partitions
        self._partition = partition
        self._note = note

        # pandas dataframe
        self._job_list = job_list

        self._input_file = file_location

    @classmethod
    def fromfile(cls, input_file):
        swf_version = None
        computer_name = None
        location = None
        acknowledge = None
        conversion = None
        max_jobs = None
        max_records = None
        preemption = None
        unix_start_time = None
        time_zone_string = None
        start_time = None
        end_time = None
        max_nodes = None
        max_procs = None
        max_runtime = None
        max_memory = None
        allow_overuse = None
        max_queues = None
        queues = None
        queue = None
        max_partitions = None
        partitions = None
        partition = None
        note = None

        # parse header file
        with open(input_file, "r") as file:
            for line_num, line in enumerate(file):
                if line[0] == ";":
                    # this is a comment line. Then we remove the first character.
                    # TODO there is a more elegant way to do this in python.

                    try:
                        field, value = line[1:].rstrip("\n").split(": ")
                    except ValueError:
                        # this is an empty comment line, so we do nothing
                        continue

                    # remove white spaces
                    field = field.lstrip().rstrip()
                    value = value.rstrip().lstrip()
                    if field == "Version":
                        swf_version = value
                    elif field == "Computer":
                        computer_name = value
                    elif field == "Installation":
                        location = value
                    elif field == "Information":
                        information = value
                    elif field == "Acknowledge":
                        acknowledge = value
                    elif field == "Conversion":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        conversion = value
                    elif field == "UnixStartTime":
                        try:
                            unix_start_time = int(value)
                        except ValueError:
                            print(f"[WARNING] Failed to cast to int {field} with value '{value}' in line {line_num}. "
                                  f"Ignoring it.")
                    elif field == "TimeZone":
                        # this field is ignored in current versions of SWF
                        pass
                    elif field == "TimeZoneString":
                        time_zone_string = value
                    elif field == "StartTime":
                        start_time = parse_time(value)
                    elif field == "EndTime":
                        end_time = parse_time(value)
                    elif field == "MaxJobs":
                        try:
                            max_jobs = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxRecords":
                        try:
                            max_records = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxNodes":
                        try:
                            max_nodes = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxProcs":
                        try:
                            max_procs = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "Preemption":
                        if value.casefold() in ["yes", "no", "double", "ts"]:
                            preemption = value
                        else:
                            print(f"[WARNING] Value '{value}' in line {line_num} for Preemption field is incorrect. "
                                  f"Ignoring it")
                    elif field == "Queues":
                        queues = value
                    elif field == "Queue":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        pass
                    elif field == "MaxQueues":
                        try:
                            max_queues = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "Partitions":
                        partitions = value
                    elif field == "Partition":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        pass
                    elif field == "MaxPartitions":
                        try:
                            max_partitions = int(value)
                        except ValueError:
                            print(f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}."
                                  f"Ignoring it.")
                    elif field == "Note":
                        note = value
                    else:
                        print(f"[WARNING] Unrecognized field {field} in the header in line {line_num}. Ignoring it.")
                else:
                    # we stop reading the file since we are not out of the header
                    break

        job_list =  pd.read_csv(input_file, names=["job_id", "submit_time", "wait_time", "runtime", "alloc_cpus",
                                                   "avg_cpu_used", "used_memory", "requested_cpus", "requested_time",
                                                   "requested_memory", "status", "uid", "gid", "app_id", "qos_id",
                                                   "partition_id", "preeceding", "thinktime"],
                                sep="\s+", comment=";")

        return cls(swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                   unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime,
                   max_memory, allow_overuse, max_queues, queues, queue, max_partitions, partitions, partition, note,
                   job_list, input_file)


    @property
    def swf_version(self):
        return self._swf_version

    @swf_version.setter
    def swf_version(self, version):
        self._swf_version = version

    @property
    def computer(self):
        return self._computer

    @computer.setter
    def computer(self, computer):
        self._computer = computer

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        self._location = location

    @property
    def acknowledge(self):
        return self._acknowledge

    @acknowledge.setter
    def acknowledge(self, acknowledge):
        self._acknowledge = acknowledge

    @property
    def conversion(self):
        return self._conversion

    @conversion.setter
    def conversion(self, conversion):
        self._conversion = conversion

    @property
    def max_jobs(self):
        return self._max_jobs

    @max_jobs.setter
    def max_jobs(self, max_jobs):
        self._max_jobs = max_jobs

    @property
    def max_records(self):
        return self._max_records

    @max_records.setter
    def max_records(self, max_records):
        self._max_records = max_records

    @property
    def preemption(self):
        return self._preemption

    @preemption.setter
    def preemption(self, preemption):
        self._preemption = preemption

    @property
    def unix_start_time(self):
        return self._unix_start_time

    @unix_start_time.setter
    def unix_start_time(self, unix_start_time):
        self._unix_start_time = unix_start_time

    @property
    def time_zone_string(self):
        return self._time_zone_string

    @time_zone_string.setter
    def time_zone_string(self, time_zone_string):
        self._time_zone_string = time_zone_string

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        self._start_time = parse_time(start_time)

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, end_time):
        self._end_time = parse_time(end_time)

    @property
    def max_nodes(self):
        return self._max_nodes

    @max_nodes.setter
    def max_nodes(self, max_nodes):
        self._max_nodes = max_nodes

    @property
    def max_procs(self):
        return self._max_procs

    @max_procs.setter
    def max_procs(self, max_procs):
        self._max_procs = max_procs

    @property
    def max_runtime(self):
        return self._max_runtime

    @max_runtime.setter
    def max_runtime(self, max_runtime):
        self._max_runtime = max_runtime

    @property
    def max_memory(self):
        return self._max_memory

    @max_memory.setter
    def max_memory(self, max_memory):
        self._max_memory = max_memory

    @property
    def allow_overuse(self):
        return self._allow_overuse

    @allow_overuse.setter
    def allow_overuse(self, allow_overuse):
        self._allow_overuse = allow_overuse

    @property
    def max_queues(self):
        return self._max_queues

    @max_queues.setter
    def max_queues(self, max_queues):
        self._max_queues = max_queues

    @property
    def queues(self):
        # TODO how to set multiple lines of comment here
        return self._queues

    @queues.setter
    def queues(self, queues):
        # TODO how to set multiple lines of comment here
        self._queues = queues

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, queue):
        self._queue = queue

    @property
    def max_partitions(self):
        return self._max_partitions

    @max_partitions.setter
    def max_partitions(self, max_partitions):
        self._max_partitions = max_partitions

    @property
    def partitions(self):
        return self._partition

    @partitions.setter
    def partitions(self, partition):
        self._partition = partition

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, note):
        self._note = note

    @property
    def input_file(self):
        return self._input_file

    @input_file.setter
    def input_file(self, input_file):
        self._input_file = input_file

    @property
    def job_list(self) -> pd.DataFrame:
        return self._job_list

    @job_list.setter
    def job_list(self, job_list):
        self._job_list = job_list

    def add_job(self, submit_time, runtime, alloc_cpus, uid, requested_time, wait_time = -1, avg_cpu_used = -1,
                used_memory= -1, requested_cpus= -1, requested_memory= -1, status= -1, gid= -1, app_id= -1, qos_id= -1,
                partition_id= -1, preeceding = -1, thinktime= -1):
        # this is to preserve the submitting order
        sub_before = self.job_list[self._job_list["submit_time"] <= submit_time]
        sub_after = self.job_list[self._job_list["submit_time"] > submit_time]
        # shift by 1 those that were submitted after than the job that it is being added
        # TODO: WE DONT NEED TO KEEP JOBID IF WE HAVE INDICES
        sub_after.loc[:, "job_id"] += 1
        new_job_id = len(sub_before)+1

        # 1 element data frame
        single_job = pd.DataFrame([[new_job_id, submit_time, wait_time, runtime, alloc_cpus, avg_cpu_used, used_memory,
                                   requested_cpus, requested_time, requested_memory, status, uid, gid, app_id, qos_id,
                                   partition_id, preeceding, thinktime]],
                                  columns=["job_id", "submit_time", "wait_time", "runtime", "alloc_cpus",
                                           "avg_cpu_used", "used_memory", "requested_cpus", "requested_time",
                                           "requested_memory", "status", "uid", "gid", "app_id", "qos_id",
                                           "partition_id", "preeceding", "thinktime"])

        print(f"Adding job to id {len(sub_before)}")
        self.job_list = pd.concat([sub_before, single_job, sub_after])

    # output to file!
    def to_bsc_swf(self, output_filename: str, ncores: int = 48) -> None:
        """
        Outputs a file with only jobs, without header, with commas for each of the job's attributes.

        Moreover, bscswf expects username, not uid, partition name, not id, and also qos name, not qos_id.

        Then we adapt users by adding a "u" in the begging of the string. E.g. uid 1001

        :param output_filename: name of the (bsc) swf file
        :param ncores: number of cores per NODE. This is because the simulator does not support running jobs less than
        the number of cores.
        :return: None
        """
        self.job_list["username"] = self.job_list["uid"].apply(lambda x: "u"+str(x))
        self.job_list["qos"] = self.job_list["qos_id"].apply(lambda x: "q" + str(x))
        self.job_list["partition"] = self.job_list["partition_id"].apply(lambda x: "p" + str(x))

        # select rows with more than nncores if allocated or requested cpus are more than 48
        more_nncores = self.job_list[(self.job_list["alloc_cpus"] >= 48) | (self.job_list["requested_cpus"] >= 48)]

        # set jobnids to match rows
        more_nncores.loc[:, "job_id"] = range(1, len(more_nncores) + 1)

        # output to bsc's format.
        more_nncores.to_csv(output_filename, columns=["job_id", "submit_time", "wait_time", "runtime", "alloc_cpus",
                                                   "avg_cpu_used", "used_memory", "requested_cpus", "requested_time",
                                                   "requested_memory", "status", "username", "gid", "app_id", "qos",
                                                   "partition", "preeceding", "thinktime"], index=False, header=False)
